(window.webpackJsonp=window.webpackJsonp||[]).push([[20],{742:function(n,t,r){"use strict";var o=r(3),e=r(743);o({target:"String",proto:!0,forced:r(744)("link")},{link:function(n){return e(this,"a","href",n)}})},743:function(n,t,r){var o=r(6),e=r(42),c=r(23),f=/"/g,w=o("".replace);n.exports=function(n,t,r,o){var l=c(e(n)),v="<"+t;return""!==r&&(v+=" "+r+'="'+w(c(o),f,"&quot;")+'"'),v+">"+l+"</"+t+">"}},744:function(n,t,r){var o=r(5);n.exports=function(n){return o((function(){var t=""[n]('"');return t!==t.toLowerCase()||t.split('"').length>3}))}}}]);