(window.webpackJsonp=window.webpackJsonp||[]).push([[110],{1158:function(t,e,d){"use strict";d.r(e);var o={props:{dataset:{type:Object,required:!0}},data:function(){return{selectedMetadataItem:void 0}},methods:{onShowMetadata:function(t){this.selectedMetadataItem=t},resetMetadataItem:function(){this.selectedMetadataItem=void 0}}},n=d(66),component=Object(n.a)(o,(function(){var t=this,e=t._self._c;return e("results-list",{attrs:{dataset:t.dataset,"metadata-item":t.selectedMetadataItem},on:{closeMetadata:t.resetMetadataItem},scopedSlots:t._u([{key:"record",fn:function(d){var o=d.record;return[e("record-text-2-text",{attrs:{dataset:t.dataset,record:o},on:{onShowMetadata:t.onShowMetadata}})]}}])})}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{ResultsList:d(958).default})}}]);