(window.webpackJsonp=window.webpackJsonp||[]).push([[116],{1167:function(e,t,r){"use strict";r.r(t);var o={props:{dataset:{type:Object,required:!0},currentMetric:{type:String}},data:function(){return{sidebarItems:[{id:"explore",tooltip:"Exploration",icon:"exploration",group:"Mode",action:"change-view-mode",relatedMetrics:["progress","stats"]},{id:"annotate",tooltip:"Hand labeling",icon:"hand-labeling",group:"Mode",action:"change-view-mode",relatedMetrics:["progress","stats"]},{id:"progress",tooltip:"Progress",icon:"progress",action:"show-metrics",group:"Metrics"},{id:"stats",tooltip:"Stats",icon:"stats",action:"show-metrics",group:"Metrics"},{id:"refresh",tooltip:"Refresh",icon:"refresh",group:"Refresh",action:"refresh"}]}},methods:{onChangeViewMode:function(e){this.$emit("change-view-mode",e)},onShowMetrics:function(e){this.$emit("show-metrics",e)}}},n=r(66),component=Object(n.a)(o,(function(){var e=this;return(0,e._self._c)("sidebar-menu",{attrs:{"current-metric":e.currentMetric,dataset:e.dataset,"sidebar-items":e.sidebarItems},on:{refresh:function(t){return e.$emit("refresh")},"show-metrics":e.onShowMetrics,"change-view-mode":e.onChangeViewMode}})}),[],!1,null,null,null);t.default=component.exports;installComponents(component,{SidebarMenu:r(819).default})}}]);