(window.webpackJsonp=window.webpackJsonp||[]).push([[26],{1012:function(e,t,n){"use strict";n.r(t);var o={props:["date"],computed:{timeDifference:function(){return(new Date).getTimezoneOffset()}}},c=n(66),component=Object(c.a)(o,(function(){var e=this;return(0,e._self._c)("span",[e._v(e._s(e._f("moment")(e._f("moment")(e._f("moment")(e.date,"utc"),"subtract","".concat(e.timeDifference," minutes")),"from","now")))])}),[],!1,null,null,null);t.default=component.exports}}]);