(window.webpackJsonp=window.webpackJsonp||[]).push([[125],{1137:function(n,t,e){"use strict";e.r(t);var o={mounted:function(){this.$router.push("/datasets")}},c=e(66),component=Object(c.a)(o,(function(){return(0,this._self._c)("div")}),[],!1,null,null,null);t.default=component.exports}}]);